#include "loadflow.h"

loadflow *read_loadflow(FILE *Datafile);
void decode_card_header(FILE *Datafile, int *Card_Type, loadflow *Loadflow);
bus_data *read_bus_data(FILE *Datafile, int Bus_Count);
void print_bus_data(bus_data *Bus_data);
branch_data *read_branch_data(FILE *Datafile, int Branch_Count);
void print_branch_data(branch_data *Branch_data);
loss_zone *read_loss_zone(FILE *Datafile, int Loss_Zone_Count);
void print_loss_zone(loss_zone *Loss_Zone);
interchange_data *read_interchange_data(FILE *Datafile,
                                        int Interchange_Data_Count);
void print_interchange_data(interchange_data *Interchange_Data);
tie_line *read_tie_line(FILE *Datafile, int Tie_Line_Count);
void print_tie_line(tie_line *Tie_Line);

#if 0
#define DEBUG_read_bus_data
#define DEBUG_read_branch_data
#define DEBUG_read_loss_zone
#define DEBUG_read_interchange_data
#define DEBUG_read_tie_line
#endif

loadflow *read_loadflow(FILE *Datafile)
{
   loadflow *Loadflow;
   char      Card_name[6];
   char      Buffer[LINE_LENGTH];
   int       Card_Type;

   if ((Loadflow = (loadflow *) malloc(sizeof(loadflow))) != NULL)
   {
      /* read in and junk the header of the file */
      fgets(Buffer, LINE_LENGTH, Datafile);
      printf("%s", Buffer);

      do
      {
         decode_card_header(Datafile, &Card_Type, Loadflow);

     
         switch (Card_Type)
         {
            case BUS_DATA:
               Loadflow->Bus_Data = read_bus_data(Datafile, Loadflow->Num_bus);
               fgets(Buffer, LINE_LENGTH, Datafile);
               printf("%s", Buffer);
               break;

            case BRANCH_DATA:
               Loadflow->Branch_Data = read_branch_data(Datafile,
                                                        Loadflow->Num_branch);
               fgets(Buffer, LINE_LENGTH, Datafile);
               printf("%s", Buffer);
               break;

            case LOSS_ZONE:
               Loadflow->Loss_Zone = read_loss_zone(Datafile,
                                                    Loadflow->Num_loss_zone);
               fgets(Buffer, LINE_LENGTH, Datafile);
               printf("%s", Buffer);
               break;

            case INTERCHANGE_DATA:
               Loadflow->Interchange_Data = read_interchange_data(Datafile,
                                               Loadflow->Num_interchange);
               fgets(Buffer, LINE_LENGTH, Datafile);
               printf("%s", Buffer);
            break;

         case TIE_LINE:
               Loadflow->Tie_Line = read_tie_line(Datafile,
                                    Loadflow->Num_tie_line);
               fgets(Buffer, LINE_LENGTH, Datafile);
               printf("%s", Buffer);
               break;

            default:
               break;
         }
      } while (Card_Type != END_OF_DATA);
   }
   else
      perror("read_loadflow()");

   return(Loadflow);
}

bus_data *read_bus_data(FILE *Datafile, int Bus_Count)
{
   bus_data *Bus_Data;
   int       i;
   int       Bus_Number;
   int       Sequence_num;

   if ((Bus_Data = (bus_data *) malloc(sizeof(bus_data) * Bus_Count)) != NULL)
   {
      for (i =0; i < Bus_Count; i++)
      {
         fscanf(Datafile, "%4d %12[A-Z 0-9.] %2d%3d %2d %7lf%7lf%9lf%10lf%8lf\
%8lf %7lf%6lf%8lf%8lf%8lf%8lf %*4d %4d%*c",
                &Bus_Data[i].Num,
                Bus_Data[i].Name,
                &Bus_Data[i].Area,
                &Bus_Data[i].Loss_Zone,
                &Bus_Data[i].Type,
                &Bus_Data[i].Final_Voltage,
                &Bus_Data[i].Final_Degree,
                &Bus_Data[i].Load_Re,
                &Bus_Data[i].Load_Im,
                &Bus_Data[i].Generation_Re,
                &Bus_Data[i].Generation_Im,
                &Bus_Data[i].Base_KV,
                &Bus_Data[i].Desired_Voltage,
                &Bus_Data[i].Limits_Max,
                &Bus_Data[i].Limits_Min,
                &Bus_Data[i].Impedience_Re,
                &Bus_Data[i].Impedience_Im,
                &Sequence_num);
#ifdef DEBUG_read_bus_data
         printf("Sequence Number = %d\n", Sequence_num);

         print_bus_data(&(Bus_Data[i]));
#endif
         if (Sequence_num != (i + 1))
         {
            printf("read_bus_data(): error reading record %d\n", i + 1);
            exit(1);
         }
      }
   }
   else
      perror("read_Bus_data()");

   return(Bus_Data);
}

void print_bus_data(bus_data *Bus_data)
{
   printf("Data of Bus %d:\n", Bus_data->Num);
   printf("   Name = [%s]\n", Bus_data->Name);
   printf("   Area = %d\n", Bus_data->Area);
   printf("   Loss Zone = %d\n", Bus_data->Loss_Zone);
   printf("   Type = %d\n", Bus_data->Type);
   printf("   Final Voltage and angle = %lf (%lf)\n",
          Bus_data->Final_Voltage, Bus_data->Final_Degree);
   printf("   Load = %lf + j %lf\n",
          Bus_data->Load_Re, Bus_data->Load_Im);
   printf("   Generation = %lf + j %lf\n",
          Bus_data->Generation_Re, Bus_data->Generation_Im);
   printf("   Base KV = %lf\n", Bus_data->Base_KV);
   printf("   Desired Voltage = %lf\n", Bus_data->Desired_Voltage);
   printf("   Bus Limits:\n");
   printf("      Max = %lf, Min = %lf\n", Bus_data->Limits_Max,
          Bus_data->Limits_Min);
   printf("   Impedience = %lf + j %lf\n",
          Bus_data->Impedience_Re, Bus_data->Impedience_Im);
   printf("\n");
}

void decode_card_header(FILE *Datafile, int *Card_Type, loadflow *Loadflow)
{
   char Card_Name[10];
   char str_data[5];
   char str_follows[8];
   char str_items[6];
   char str_header[9];
   int  Number_Item;

   fscanf(Datafile, "%s %s %s %d %s %s%*c", Card_Name, str_data,
          str_follows, &Number_Item, str_items, str_header);

   printf("Card_Name = %s\n", Card_Name);
   printf("str_data = %s\n", str_data);
   printf("str_follows = %s\n", str_follows);
   printf("Number_Item = %d\n", Number_Item);
   printf("str_items = %s\n", str_items);
   printf("str_header = %s\n", str_header);
   printf("\n");

   switch (Card_Name[0])
   {
      case 'B':
         if (strlen(Card_Name) == 3)
         {
            /* assume BUS card */
            *Card_Type = BUS_DATA;
            Loadflow->Num_bus = Number_Item;
         }
         else
         {
            *Card_Type = BRANCH_DATA;
            Loadflow->Num_branch = Number_Item;
         }
         break;

      case 'E':
         *Card_Type = END_OF_DATA;
         break;

      case 'I':
         *Card_Type = INTERCHANGE_DATA;
         Loadflow->Num_interchange = Number_Item;
         break;

      case 'L':
         *Card_Type = LOSS_ZONE;
         Loadflow->Num_loss_zone = Number_Item;
         break;

      case 'T':
         *Card_Type = TIE_LINE;
         Loadflow->Num_tie_line = Number_Item;
         break;

      default:
         *Card_Type = UNKNOWN;
         printf("decode_card_header(): unknown card type\n");
         break;
   }
}

branch_data *read_branch_data(FILE *Datafile, int Branch_Count)
{
   branch_data *Branch_Data;
   int          i;
   int          Sequence_num;

   if ((Branch_Data = (branch_data *) malloc(sizeof(branch_data)
                                             * Branch_Count)) != NULL)
  {
      for (i = 0; i < Branch_Count; i++)
      {
         fscanf(Datafile, "%4d %4d %2d%2d %1d %1d%10lf%10lf%10lf%6lf%6lf%6lf\
%4d %1d  %6lf %7lf%7lf%7lf %6lf %7lf%7lf  %4d%*c",
                &Branch_Data[i].Tab_bus,
                &Branch_Data[i].Impedience_bus,
                &Branch_Data[i].Area,
                &Branch_Data[i].Zone, 
                &Branch_Data[i].Circuit,
                &Branch_Data[i].Type,
                &Branch_Data[i].Impedience_Re,
                &Branch_Data[i].Impedience_Im,
                &Branch_Data[i].Charging,
                &(Branch_Data[i].Line_Rating[1]),
                &(Branch_Data[i].Line_Rating[2]),
                &(Branch_Data[i].Line_Rating[3]),
                &Branch_Data[i].Control_bus,
                &Branch_Data[i].Side,
                &Branch_Data[i].Transformer_final_ratio,
                &Branch_Data[i].Transformer_final_angle,
                &Branch_Data[i].Tap_Phase_shift_Max,
                &Branch_Data[i].Tap_Phase_shift_Min,
                &Branch_Data[i].Step_size,
                &Branch_Data[i].Voltage_Limit_Max,
                &Branch_Data[i].Voltage_Limit_Min,
                &Sequence_num);

#ifdef DEBUG_read_branch_data
         printf("Sequence Number = %d\n", Sequence_num);

         print_branch_data(&(Branch_Data[i]));
#endif
         if (Sequence_num != (i + 1))
         {
            printf("read_branch_data(): error reading record %d\n", i + 1);
            exit(1);
         }
      }
   }
   else
      perror("read_branch_data()");

   return(Branch_Data);
}

void print_branch_data(branch_data *Branch_data)
{
   printf("Branch Data:\n");
   printf("   Tab bus = %d\n", Branch_data->Tab_bus);
   printf("   Impedience bus = %d\n", Branch_data->Impedience_bus);
   printf("   Area = %d\n", Branch_data->Area);
   printf("   Zone = %d\n", Branch_data->Zone);
   printf("   Circuit = %d\n", Branch_data->Circuit);
   printf("   Type = %d\n", Branch_data->Type);
   printf("   Impedience = %lf + j %lf\n",
          Branch_data->Impedience_Re, Branch_data->Impedience_Im);
   printf("   Charging = %lf\n", Branch_data->Charging);
   printf("   Line rating 1 = %lf\n", Branch_data->Line_Rating[1]);
   printf("   Line rating 2 = %lf\n", Branch_data->Line_Rating[2]);
   printf("   Line rating 3 = %lf\n", Branch_data->Line_Rating[3]);
   printf("   Control bus = %d\n", Branch_data->Control_bus);
   printf("   Side = %d\n", Branch_data->Side);
   printf("   Transformer Final ratio and angle = %lf (%lf)\n",
          Branch_data->Transformer_final_ratio,
          Branch_data->Transformer_final_angle);
   printf("   Tab or Phase shifter Limits:\n");
   printf("      Max = %lf, Min = %lf, Step size = %lf\n",
          Branch_data->Tap_Phase_shift_Max, Branch_data->Tap_Phase_shift_Min,
          Branch_data->Step_size);
   printf("   Voltage Limits:\n");
   printf("      Max = %lf, Min = %lf\n", Branch_data->Voltage_Limit_Max,
          Branch_data->Voltage_Limit_Min);
   printf("\n");
}

loss_zone *read_loss_zone(FILE *Datafile, int Loss_Zone_Count)
{
   loss_zone *Loss_Zone;
   int        i;
   int        Sequence_num;

   if ((Loss_Zone = (loss_zone *) malloc(sizeof(loss_zone)
                                         * Loss_Zone_Count)) != NULL)
  {
      for (i = 0; i < Loss_Zone_Count; i++)
      {
         fscanf(Datafile, "%3d %12[A-Z a-z0-9.] %4d%*c", &Loss_Zone[i].Zone,
                Loss_Zone[i].Name, &Sequence_num);
#ifdef DEBUG_read_loss_zone
         printf("Sequence Number = %d\n", Sequence_num);

         print_loss_zone(&(Loss_Zone[i]));
#endif
         if (Sequence_num != (i + 1))
         {
            printf("read_loss_zone(): error reading record %d\n", i + 1);
            exit(1);
         }

      }
  }
  else
     perror("read_loss_zone()");

  return(Loss_Zone);
}

void print_loss_zone(loss_zone *Loss_Zone)
{
   printf("Loss Zone:\n");
   printf("   Zone = %d\n", Loss_Zone->Zone);
   printf("   Name = [%s]\n", Loss_Zone->Name);
   printf("\n");
}

interchange_data *read_interchange_data(FILE *Datafile,
                                        int Interchange_Data_Count)
{
   interchange_data *Interchange_Data;
   int        i;
   int        Sequence_num;

   if ((Interchange_Data = (interchange_data *) malloc(sizeof(interchange_data)
                                                       * Interchange_Data_Count)
       ) != NULL)
  {
      for (i = 0; i < Interchange_Data_Count; i++)
      {
         fscanf(Datafile, "%2d %4d %12[A-Z a-z0-9.]%8lf%7lf%*c%*C\
%6[A-Z a-z0-9.]  %30[A-Z a-z0-9.] %4d%*c",
                &Interchange_Data[i].Area,
                &Interchange_Data[i].Bus_num,
                Interchange_Data[i].Bus_Name,
                &Interchange_Data[i].Scheduled_Export,
                &Interchange_Data[i].Tolerance,
                Interchange_Data[i].Area_Code,
                Interchange_Data[i].Area_Name, &Sequence_num);

#ifdef DEBUG_read_interchange_data
         printf("Sequence Number = %d\n", Sequence_num);

         print_interchange_data(&(Interchange_Data[i]));
#endif

         if (Sequence_num != (i + 1))
         {
            printf("read_interchange_data(): error reading record %d\n", i + 1);
            exit(1);
         }

      }
  }
  else
     perror("read_interchange_data()");

  return(Interchange_Data);
}

void print_interchange_data(interchange_data *Interchange_Data)
{
   printf("Interchange Data:\n");
   printf("   Area = %d\n", Interchange_Data->Area);
   printf("   Bus:\n");
   printf("      Number = %d\n", Interchange_Data->Bus_num);
   printf("      Name = [%s]\n", Interchange_Data->Bus_Name);
   printf("   Scheduled Export = %lf\n", Interchange_Data->Scheduled_Export);
   printf("   Tolerance = %lf\n", Interchange_Data->Tolerance);
   printf("   Area:\n");
   printf("      Code = [%s]\n", Interchange_Data->Area_Code);
   printf("      Name = [%s]\n", Interchange_Data->Area_Name);
   printf("\n");
}
 
tie_line *read_tie_line(FILE *Datafile, int Tie_Line_Count)
{
   tie_line *Tie_Line;
   int        i;
   int        Sequence_num;

   if ((Tie_Line = (tie_line *) malloc(sizeof(tie_line)
                                         * Tie_Line_Count)) != NULL)
  {
      for (i = 0; i < Tie_Line_Count; i++)
      {
         fscanf(Datafile, "%4d %2d %4d %2d %1d %4d%*c",
                &Tie_Line[i].Meter_Bus_num,
                &Tie_Line[i].Meter_Bus_Area,
                &Tie_Line[i].Non_Meter_Bus_num,
                &Tie_Line[i].Non_Meter_Bus_Area,
                &Tie_Line[i].Circuit,
                &Sequence_num);
#ifdef DEBUG_read_tie_line
         printf("Sequence Number = %d\n", Sequence_num);

         print_tie_line(&(Tie_Line[i]));
#endif
         if (Sequence_num != (i + 1))
         {
            printf("read_tie_line(): error reading record %d\n", i + 1);
            exit(1);
         }

      }
  }
  else
     perror("read_tie_line()");

  return(Tie_Line);
}

void print_tie_line(tie_line *Tie_Line)
{
   printf("Tie Line:\n");
   printf("   Meter Bus:\n");
   printf("      Number = %d\n", Tie_Line->Meter_Bus_num);
   printf("      Area = %d\n", Tie_Line->Meter_Bus_Area);
   printf("   Non-meter Bus:\n");
   printf("      Number = %d\n", Tie_Line->Non_Meter_Bus_num);
   printf("      Area = %d\n", Tie_Line->Non_Meter_Bus_Area);
   printf("   Circuit = %d\n", Tie_Line->Circuit);
   printf("\n");
}

main()
{
   FILE *Datafile;

   if ((Datafile = fopen("ieee50.dat", "r")) != NULL)
   {
      read_loadflow(Datafile);
      close(Datafile);
   }
   else
      perror("load flow reader test program");

}
